class DankCordException(BaseException):
    pass


class InvalidToken(DankCordException):
    pass


class MissingPermissions(DankCordException):
    pass


class UnknownChannel(DankCordException):
    pass


class NoCommands(DankCordException):
    pass


class DataAccessFailure(DankCordException):
    pass


class InvalidFormBody(DankCordException):
    pass


from typing import Literal, Optional, Union
from rich import print
from re import findall
from datetime import datetime

class Config:
    def __init__(
        self,
        token: str,
        channel_id: Optional[int],
        dm_mode: bool = False,
        resource_intensivity: Literal["DISK", "MEM"] = "MEM",
    ) -> None:
        if not token:
            raise ValueError("No token provided.")
        if not channel_id:
            raise ValueError("No channel_id provided.")

        assert isinstance(token, str), "Bot token must be of type str."
        assert isinstance(channel_id, int), "Channel ID must be of type int."
        assert resource_intensivity.upper() in ("DISK", "MEM"), "Resource intensivity option must be either DISK or MEM."

        self.token: str = token
        self.channel_id: int = channel_id
        self.dm_mode: bool = dm_mode
        self.resource_intensivity: str = resource_intensivity.upper()


class Cache:
    def __init__(self) -> None:
        """
        Holds relevant websocket message events.
        """
        self.nonce_message_map = {}
        self.interaction_create = []
        self.interaction_success = []
        self.raw_message_updates = []
        self.message_create = {}
        self.message_updates = {}

    def clear(self, nonce):
        if nonce in self.nonce_message_map:
            relevant_id = self.nonce_message_map[nonce]
            del self.nonce_message_map[nonce]
            if relevant_id in self.message_updates:
                del self.message_updates[relevant_id]
        if nonce in self.interaction_create:
            self.interaction_create.remove(nonce)
        if nonce in self.interaction_success:
            self.interaction_success.remove(nonce)
        if nonce in self.message_create:
            del self.message_create[nonce]
        self.raw_message_updates = []


class Author:
    """
    A class that represents the author of some context.
    """

    def __init__(self, data: dict) -> None:
        self.name: Optional[str] = data.get("name", None)
        self.discriminator: Optional[str] = data.get("discriminator", None)
        self.icon_url: Optional[str] = data.get("icon_url", None)
        self.id: Optional[int] = int(data.get("id", 0))
    
    def __repr__(self) -> str:
        return f"{self.name}#{self.discriminator}"

class ActionRow:
    """
    Represents an ActionRow.
    """

    def __init__(self, data: dict, message_id: Union[str, int]):
        message_id = str(message_id) # type: ignore
        self.components: list[Button | Dropdown] = [
            Button(i, message_id) if i["type"] == 2 else Dropdown(i, message_id) for i in data["components"]
        ]
    

class DropdownOption:
    """
    Represents a Dropdown Option.
    """

    def __init__(self, data: dict) -> None:
        self.label: Optional[str] = data.get("label", None)
        self.value: Optional[str] = data.get("value", None)
        self.value: str = data.get("value", None)


class Dropdown:
    """
    Represents a Dropdown component.
    """

    def __init__(self, data: dict, message_id: str) -> None:
        self.message_id: str = message_id
        self.type: int = 3
        self.custom_id: Optional[int] = data.get("custom_id", None)
        self.options: list[DropdownOption] = [DropdownOption(child) for child in data["options"]]

    # TODO: Make a choose function


class Button:
    """
    Represents a button from the Discord Bot UI Kit.
    """

    def __init__(self, data: dict, message_id: str) -> None:
        self.message_id: str = message_id
        self.type: int = 2
        self.emoji: Optional[Emoji] = Emoji(data["emoji"]) if "emoji" in data.keys() else None
        self.label: Optional[str] = data.get("label", None)
        self.disabled: Optional[bool] = data.get("disabled", False)
        self.custom_id: Optional[str] = data.get("custom_id", None)


class Emoji:
    """
    Represents a custom emoji.
    """

    def __init__(self, data: dict) -> None:
        self.name: Optional[str] = data.get("name", None)
        self.id: Optional[int] = data.get("id", None)


class EmbedFooter:
    """
    Represents an embed footer.
    """

    def __init__(self, data: dict) -> None:
        self.text: Optional[str] = data.get("text", None)
        self.icon_url: Optional[str] = data.get("icon_url", None)
        self.proxy_icon_url: Optional[str] = data.get("proxy_icon_url", None)


class Embed:
    """
    Represents a Discord embed.
    """

    def __init__(self, data: dict) -> None:
        self.title: str = data.get("title", None)
        self.description: str = data.get("description", None)
        self.authorName : str = data.get("author", {}).get("name", None)
        self.url: str = data.get("url", None)
        self.author: Optional[Author] = Author(data["author"]) if "author" in data else None
        self.footer: Optional[EmbedFooter] = EmbedFooter(data["footer"]) if "footer" in data else None


class Message:
    """
    Represents a message from Discord.
    """

    def __init__(self, data: dict) -> None:
        self.data: dict = data
        self.author: Optional[Author] = Author(data["author"]) if "author" in data else None
        self.content: Optional[str] = data.get("content", None)
        self.nonce: Optional[str] = data.get("nonce", None)
        self.id: Optional[int] = int(data.get("id", 0))
        self.timestamp: Optional[str] = data.get("timestamp", None)
        self.channel_id: Optional[int] = int(data.get("channel_id", 0))
        self.embeds: list[Embed] = [Embed(i) for i in data.get("embeds", [])]
        self.components: list[ActionRow] = [ActionRow(i, self.id) for i in data.get("components", [])]
        self.buttons: list[Button] = [
            item for component in self.components for item in component.components if isinstance(item, Button)
        ]
        self.dropdowns: list[Dropdown] = [
            item for component in self.components for item in component.components if isinstance(item, Dropdown)
        ]

class Bot:
    """
    Represents the bot account.
    """
    
    def __init__(self, data: dict) -> None:
        self.username: str = data["d"]["user"]["username"]
        self.id: int = int(data["d"]["user"]["id"])
        self.discriminator: int = int(data["d"]["user"]["discriminator"])
        self.email: str = data["d"]["user"]["email"]
        self.bot: str = f"{self.username}#{self.discriminator}"

class CommandResult:
    """
    Represents a class that has the result of running a certain command.
    """
    
    def __init__(self, success: bool, death: bool = False, gain: dict = {}, loss: dict = {}, cooldown: int = None) -> None:
        self.success: Optional[bool] = success
        self.death: Optional[bool] = death
        self.gain: Optional[dict] = gain
        self.loss: Optional[dict] = loss
        self.cooldown: Optional[int] = cooldown
        if self.cooldown:
            self.cooldown = round(self.cooldown, 2)

class Parser:
    """
    Extracts crucial data from the result of a command, such as coins, gain, loss anad deaths.
    """
    def beg(description: str):
        """
        Parses crucial information from the descriptions of the beg command.
        """
        gain_regex = ["\*\*⏣ [0-9]+\*\*", "[0-9]+", "\*\*(.*?)\*\*", "<(.*?)\>"]
        temp_desc = description.replace(",", "")
        death = None
        success = True
        gain = {}
        gained_coins = None
        gained_item = None
        try:
            gained_coins = int(findall(gain_regex[1], findall(gain_regex[0], temp_desc)[0])[0])
        except:
            pass
        try:
            if not gained_coins:
                gained_item = findall(gain_regex[2], temp_desc)[0]
                tempstring = "<" + findall(gain_regex[3], gained_item)[0] + "> "
                item = item.replace(tempstring, "")
            else:
                gained_item = findall(gain_regex[2], temp_desc)[1]
                tempstring = "<" + findall(gain_regex[3], gained_item)[0] + "> "
                gained_item = gained_item.replace(tempstring, "")
        except:
            pass
        if gained_coins:
            gain["coins"] = gained_coins
        if gained_item:
            gain["items"] = [{1: gained_item}]
        success = not gained_coins is None or not gained_item is None
        return CommandResult(success, death, gain)
    
    def search(description: str):
        """
        Parses crucial information from the descriptions of the search command.
        """
        gain_regex = ["\*\*⏣ [0-9]+\*\*", "[0-9]+"]
        gain = {}
        temp_desc = description.replace(",", "")
        item_amount = None
        item = None
        try:
            item_amount = int(findall("\*\*[0-9]+x", temp_desc)[0].replace("**", "").replace("x", ""))
            item = " ".join(findall("\*\*[0-9]+x .+\*\*", temp_desc)[0].replace("**", "").split(" ")[2:])
        except:
            pass
        death = None
        success = True
        gained_coins = None
        try:
            gained_coins = int(findall(gain_regex[1], findall(gain_regex[0], temp_desc)[0])[0])
        except:
            success = False
        if gained_coins:
             gain["coins"] = gained_coins
        if item_amount:
            gain["items"] = [{item_amount: item}]
        return CommandResult(success, death, gain)

    def common1(description: str):
        """
        Parses crucial information from the descriptions of the fish, hunt and dig commands.
        """
        temp_desc = description.replace(",", "")
        gain_regex = ["\*\*.*\*"]
        gained_items = None
        success = True
        death = None
        gain = {}
        try:
            gained_items =  " ".join(findall(gain_regex[0], temp_desc)[0].replace("**", "").split(" ")[1:])
        except:
            success = False
        if gained_items:
            gain["items"] = [{1: gained_items}]
        return CommandResult(success, death, gain)
        
    def crime(description: str):
        """
        Parses crucial information from the descriptions of the crime command.
        """
        gain_regex = ["\*\*⏣ [0-9]+\*\*", "[0-9]+", "<(.*?)\>", "\*\*(.*?)\*\*"]
        gain = {}
        temp_desc = description.replace(",", "")
        item = None
        death = None
        success = False
        gained_coins = None
        try:
            gained_coins = int(findall(gain_regex[1], findall(gain_regex[0], temp_desc)[0])[0])
        except:
            pass
        try:
            if not gained_coins:
                item = findall(gain_regex[3], temp_desc)[0]
                tempstring = "<" + findall(gain_regex[2], item)[0] + "> "
                item = item.replace(tempstring, "")
            else:
                item = findall(gain_regex[3], temp_desc)[1]
                tempstring = "<" + findall(gain_regex[2], item)[0] + "> "
                item = item.replace(tempstring, "")
        except:
            pass
        if gained_coins:
             gain["coins"] = gained_coins
        if item:
            gain["items"] = [{1: item}]
        success = not gained_coins is None or not item is None
        return CommandResult(success, death, gain)
    
    def postmemes(description: str):
        """
        Parses crucial information from the descriptions of the postmemes command.
        """
        parsed = description.replace(",", "").split("\n")[3:]
        try:
            index = parsed.index("**You Received:**")
            parsed = parsed[index:]
        except ValueError:
            pass
        gain = {}
        success = True
        death = None
        gain_regex = ["[0-9]+", "[0-9]+x"]
        if "**You Received" in description:
            for i in parsed:
                if "⏣" in i:
                    gain["coins"] = int(findall(gain_regex[0], i)[0])
                elif "<:" in i:
                    if not "items" in gain.keys():
                        gain["items"] = []
                    amount = int(findall(gain_regex[1], i)[0].replace('x', ''))
                    item = " ".join(i.split(" ")[5:])
                    gain["items"].append({amount: item})
        else:
            success = False
        return CommandResult(success, death, gain)
    
    def cooldown(description: str):
        """
        Parses crucial information from the descriptions of the cooldown indicator.
        """
        cooldown_regex = ["<(.*?)\>", "(\d+)"]
        time = datetime.fromtimestamp(int(findall(cooldown_regex[1], findall(cooldown_regex[0], description)[0])[0]))
        difference = (time - datetime.now()).total_seconds()
        return CommandResult(False, None, {}, cooldown=difference)
    
    def check_cooldown(description: str):
        """
        Checks if the returned embed is a cooldown message.
        """
        try:
            if "seconds" in description and "command" in description and "cooldown is" in description and "seconds" in description:
                return True
            else:
                return False
        except:
            return False
        
class User:
    """
    Represents a class that has the data of a user.
    """

    def __init__(self, data: dict) -> None:
        self.id: Optional[int] = int(data.get("id", 0))
        self.discriminator: Optional[int] = int(data.get("discriminator", 0))
        self.name: Optional[str] = data.get("username", None)
        self.bio: Optional[str] = data.get("bio", None)
        self.phone: Optional[str] = data.get("phone", None)
        self.email: Optional[str] = data.get("email", None)
        self.verified: bool = data.get("verified", False)


import threading, time, orjson

from rich import print
from typing import Optional
from pyloggor import pyloggor
from websocket import create_connection

from .exceptions import InvalidToken
from .Objects import Cache, Config


class GatewayInternal:
    def __init__(self) -> None:
        self.s = 0
        self.resume_gateway_url = ""


class Gateway:
    def __init__(self, config: Config, logger: pyloggor) -> None:
        logger.log(level="Debug", msg="Booting up gateway instance.")
        self.token = config.token
        self.logger = logger

        self.channel_id = config.channel_id
        self.dm_mode = config.dm_mode

        self.session_id: Optional[str] = None
        self.user_id: Optional[int] = None
        self.guild_id: Optional[int] = None

        self.internal: GatewayInternal = GatewayInternal()
        self.cache: Cache = Cache()
        self.pause = False

        self.__boot_ws()

    def heartbeat(self):
        self.logger.log(level="Debug", msg="Booted up discord heartbeat loop.")
        self.pause = False
        while True:
            if not self.pause:
                time.sleep(self.heartbeat_interval)
                self.ws.send(orjson.dumps({"op": 1, "d": None}))

    def recv_handler(self):
        try:
            event = self.ws.recv()
            data = orjson.loads(event)
            return data
        except:
            return False

    def __boot_ws(self):
        start = time.perf_counter()
        self.logger.log(level="Debug", msg="Booting up websocket client.")
        ws = create_connection("wss://gateway.discord.gg/?v=9&encoding=json")
        self.ws = ws

        hello = self.recv_handler()
        if not hello:
            self.logger.log(level="Critical", msg="Discord websocket client failed to boot.")
            return False

        self.heartbeat_interval = hello["d"]["heartbeat_interval"] / 1000
        self.logger.log(level="Debug", msg="Received gateway hello event.")

        self.ws.send(orjson.dumps({"op": 1, "d": None}))
        heartbeat_init = self.recv_handler()
        if not heartbeat_init:
            self.logger.log(level="Critical", msg="Discord heartbeat loop failed to boot.")
            return False

        if heartbeat_init["op"] == 11:
            pass
        elif heartbeat_init["op"] == 1:
            self.ws.send(orjson.dumps({"op": 1, "d": None}))
            heartbeat_init_2 = self.recv_handler()
            if not heartbeat_init_2:
                self.logger.log(level="Critical", msg="Discord heartbeat loop failed to boot.")
                return False
            if heartbeat_init_2["op"] != 11:
                self.logger.log(level="Critical", msg="Unhandled OP response while booting discord heartbeat loop.")
                return False
        else:
            self.logger.log(level="Critical", msg="Unhandled OP response while booting discord heartbeat loop.")
            return False

        threading.Thread(target=self.heartbeat).start()

        ws.send(
            orjson.dumps(
                {
                    "op": 2,
                    "d": {
                        "token": self.token,
                        "properties": {
                            "$os": "windows",
                            "$browser": "Discord",
                            "$device": "desktop",
                        },
                    },
                }
            )
        )

        identify = self.recv_handler()
        if not identify:
            raise InvalidToken("Invalid Discord account token used.") # type: ignore

        if identify["op"] != 0:
            self.logger.log(
                level="UNHANDLED DEBUG",
                msg="Unhandled case, send this to the developers.",
                extras={"identify_response": identify},
            )
            return False

        self.internal.s = identify["s"]

        end = time.perf_counter()

        if not self.dm_mode:
            self.logger.log(level="Debug", msg="Caching guild ID")
            for guild in identify["d"]["guilds"]:
                for channel in guild.get("channels", []):
                    if int(channel["id"]) == int(self.channel_id):
                        self.logger.log(level="Debug", msg="Found guild ID.")
                        self.guild_id = guild["id"]
                        break

        self.user_id = int(identify["d"]["user"]["id"])
        self.session_id = identify["d"]["session_id"]
        self.internal.resume_gateway_url = identify["d"]["resume_gateway_url"]

        thread = threading.Thread(target=self._events_listener)
        thread.daemon = True
        thread.start()
        self.logger.log(level="Info", msg=f"Connected to gateway in {round(end - start, 3)} seconds.")

    def reconnect_ws(self):
        self.logger.log(level="Info", msg="Reconnecting websocket client.")
        ws = create_connection(self.internal.resume_gateway_url)
        self.ws = ws
        self.ws.send(
            orjson.dumps({"op": 6, "d": {"token": self.token, "session_id": self.session_id, "seq": self.internal.s}})
        )

        thread = threading.Thread(target=self._events_listener)
        thread.daemon = True
        thread.start()
        self.pause = False

    def _events_listener(self):
        self.logger.log(level="Debug", msg="Events listener is now listening.")
        while True:
            event = self.recv_handler()
            if not event:
                continue
            
            try:
                if event["op"] == 0:
                    self.internal.s = event["s"]

                elif event["op"] == 7 or (event["op"] == 9 and event["d"]):
                    self.reconnect_ws()
                    return
                elif event["op"] == 1:
                    self.ws.send(orjson.dumps({"op": 1, "d": None}))
                    continue
                elif event["op"] == 9 and not event["d"]:
                    self.pause = True
                    self.__boot_ws()
                    return

                if not event["d"]:
                    continue
                if event["t"] == "MESSAGE_ACK":
                    continue

                if event["t"] not in ("MESSAGE_CREATE", "MESSAGE_UPDATE", "INTERACTION_CREATE", "INTERACTION_SUCCESS"):
                    continue
                
                if event["t"] == "INTERACTION_CREATE":
                    self.cache.interaction_create.append(event["d"]["nonce"])
                elif event["t"] == "INTERACTION_SUCCESS":
                    self.cache.interaction_success.append(event["d"]["nonce"])
                elif event["t"] == "MESSAGE_CREATE":
                    try:
                        self.cache.message_create[event["d"]["nonce"]] = event["d"]
                        self.cache.nonce_message_map[event["d"]["nonce"]] = event["d"]["id"]
                    except:
                        pass
                elif event["t"] == "MESSAGE_UPDATE":
                    if event["d"]["id"] not in self.cache.message_updates:
                        self.cache.message_updates[event["d"]["id"]] = []
                    self.cache.message_updates[event["d"]["id"]].append(event["d"])
                    self.cache.raw_message_updates.append(event["d"])
                else:
                    print("----------------- DEBUG START -----------------")
                    print(event)
                    print("----------------- DEBUG END -----------------")
            except Exception as e:
                self.logger.log(level="Error", msg=f"_events_listener function in gateway.py: {e}.")


import datetime, json, time
import requests

from typing import Optional, Literal, Union, Callable
from pyloggor import pyloggor
from random import randint

from .Objects import Config, Message, Parser
from .gateway import Gateway
from .api import API

class Core(API):
    def __init__(self,
        /,
        config: Config,
        commands_data: dict,
        guild_id : Optional[int],
        session_id : Optional[str],
        logger: pyloggor,
        gateway: Gateway
    ) -> None:
        self.token = config.token
        self.channel_id = config.channel_id
        self.dm_mode = config.dm_mode
        self.resource_intensivity = config.resource_intensivity
        self.session_id = session_id
        self.commands_data = commands_data
        self.guild_id = guild_id
        self.logger = logger
        self.ws_cache = {}
        self.gateway = gateway

    def _get_command_info(self, name: str) -> dict:
        """Retuns information about a given command.
        
        Parameters
        --------
        name: str
            The name of the command.
            
        Returns
        --------
        information: dict
        """
        if self.resource_intensivity == "MEM":
            return self.commands_data.get(name, {})
        else:
            return json.load(open(f"{self.channel_id}_commands.json", "r+")).get(name, {})

    def wait_for(
        self,
        event: Literal["MESSAGE_CREATE", "MESSAGE_UPDATE", "INTERACTION_CREATE", "INTERACTION_SUCCESS"],
        check: Optional[Callable[..., bool]] = None,
        timeout: float = 10
    ) -> Optional[Union[Message, bool]]:
        """
        Waits for a WebSocket event to be dispatched.

        This could be used to wait for a message to be sent or a message to be edited,
        or even to confirm an interaction being created or successful.

        The `timeout` parameter specifies how long to wait for until the desired event
        is dispatched; if the event was not dispatched before the timeout duration is over,
        it returns `None`.

        This function returns the **first event that meets the requirements**.

        Example
        ---------

        Waiting for a message to be sent:

            def DankMemerShop():
                def check(message: Message):
                    # The author ID is the ID of Dank Memer
                    return message.author.id == 270904126974590976 and "shop" in message.embeds[0].title.lower()
                
                message: Message = bot.wait_for("MESSAGE_CREATE", check = check)

        Parameters
        ------------
        event: str
            The event name.
        check: Optional[Callable[..., `bool`]]
            A predicate to check what to wait for. The arguments must meet the
            parameters of the event being waited for.
        timeout: Optional[`float`]
            The number of seconds to wait before timing out and returning `None`.

        Returns
        --------
        Optional[Union[Message, bool]]
            Returns the Message object, or a boolean.
        """
        limit = time.time() + timeout
        if check is None:
            def _check(*args):
                return True
            check = _check

        while time.time() < limit:
            try:
                cache = self.gateway.cache            
                if event == "INTERACTION_CREATE":
                    if check(cache.interaction_create[-1]):
                        return True
                if event =="INTERACTION_SUCCESS":
                    if check(cache.interaction_success[-1]):
                        return True

                if event == "MESSAGE_CREATE":
                    value = list(cache.message_create.values())[-1]
                    _msg = Message(value)
                    if check(_msg) is True:
                        return _msg
                if event == "MESSAGE_UPDATE":
                    while len(cache.raw_message_updates) == 0:
                        continue
                    _msg = Message(cache.raw_message_updates[-1])
                    if check(_msg) is True:
                        return _msg
            except:
                pass
        return None

    # Raw commands
    def fish(self, retry_attempts:int = 3, timeout:int = 10):
        """
        Runs the `fish` command.
        
        Arguments
        --------
        retry_attempts: `int`
            The amount of times to retry when executing the command fails.
        timeout: `int`
            The time it waits for to confirm whether a command was ran or not.
        
        Raises
        --------
        InvalidComponent
            invalidcomponentdescription
        NonceTimeout
            Couldn't get the nonce from the Websocket.
        UnknownChannel
            Bot doesn't have access to that channel.

        Returns
        --------
        message: Optional[`Message`]
        """
        cmd : Message = self.run_command("fish", retry_attempts, timeout)
        if Parser.check_cooldown(cmd.embeds[0].description):
            return Parser.cooldown(cmd.embeds[0].description)
        return Parser.common1(cmd.embeds[0].description)
        
    def hunt(self, retry_attempts:int = 3, timeout:int = 10):
        """
        Runs the `hunt` command.
        
        Arguments
        --------
        retry_attempts: `int`
            The amount of times to retry when executing the command fails.
        timeout: `int`
            The time it waits for to confirm whether a command was ran or not.
        
        Raises
        --------
        InvalidComponent
            invalidcomponentdescription
        NonceTimeout
            Couldn't get the nonce from the Websocket.
        UnknownChannel
            Bot doesn't have access to that channel.

        Returns
        --------
        message: Optional[`Message`]
        """
        cmd : Message = self.run_command("hunt", retry_attempts, timeout)
        if Parser.check_cooldown(cmd.embeds[0].description):
            return Parser.cooldown(cmd.embeds[0].description)
        return Parser.common1(cmd.embeds[0].description)
        
    def dig(self, retry_attempts:int = 3, timeout:int = 10):
        """
        Runs the `dig` command.
        
        Arguments
        --------
        retry_attempts: `int`
            The amount of times to retry when executing the command fails.
        timeout: `int`
            The time it waits for to confirm whether a command was ran or not.
        
        Raises
        --------
        InvalidComponent
            invalidcomponentdescription
        NonceTimeout
            Couldn't get the nonce from the Websocket.
        UnknownChannel
            Bot doesn't have access to that channel.

        Returns
        --------
        message: Optional[`Message`]
        """
        cmd : Message = self.run_command("dig", retry_attempts, timeout)
        if Parser.check_cooldown(cmd.embeds[0].description):
            return Parser.cooldown(cmd.embeds[0].description)
        return Parser.common1(cmd.embeds[0].description)
        
    def beg(self, retry_attempts:int = 3, timeout:int = 10):
        """
        Runs the `beg` command.
        
        Arguments
        --------
        retry_attempts: `int`
            The amount of times to retry when executing the command fails.
        timeout: `int`
            The time it waits for to confirm whether a command was ran or not.
        
        Raises
        --------
        InvalidComponent
            invalidcomponentdescription
        NonceTimeout
            Couldn't get the nonce from the Websocket.
        UnknownChannel
            Bot doesn't have access to that channel.

        Returns
        --------
        message: Optional[`Message`]
        """
        cmd: Message = self.run_command("beg", retry_attempts, timeout)
        if Parser.check_cooldown(cmd.embeds[0].description):
            return Parser.cooldown(cmd.embeds[0].description)
        return Parser.beg(cmd.embeds[0].description)
    
    # Button commands
    def search(self, retry_attempts:int = 3, timeout:int = 10, location_index:Literal[1, 2, 3, "random"] = 2):
        """
        Runs the `search` command.
        
        Arguments
        --------
        retry_attempts: `int`
            The amount of times to retry when executing the command fails.
        timeout: `int`
            The time it waits for to confirm whether a command was ran or not.
        location_index: `Literal[1, 2, 3, "random"]`
            The location to search for.
        Raises
        --------
        InvalidComponent
            invalidcomponentdescription
        NonceTimeout
            Couldn't get the nonce from the Websocket.
        UnknownChannel
            Bot doesn't have access to that channel.

        Returns
        --------
        message: Optional[`Message`]
        """
        _location = None
        if location_index not in [1, 2, 3, "random"]:
            _location = "random"
        _location = location_index if location_index in [1, 2, 3, "random"] else randint(1, 3)
        cmd : Message = self.run_command("search", retry_attempts, timeout)
        if Parser.check_cooldown(cmd.embeds[0].description):
            return Parser.cooldown(cmd.embeds[0].description)
        self.click(cmd.buttons[_location - 1])
        def check(message):
            return message.channel_id == int(self.channel_id) and message.author.id == 270904126974590976 and "searched" in message.embeds[0].authorName
        cmd = self.wait_for("MESSAGE_UPDATE", check=check)
        return Parser.search(cmd.embeds[0].description)
        

    def crime(self, retry_attempts: int = 3, timeout:int = 10, location_index:Literal[1, 2, 3, "random"] = 2):
        """
        Runs the `crime` command.
        
        Arguments
        --------
        retry_attempts: `int`
            The amount of times to retry when executing the command fails.
        timeout: `int`
            The time it waits for to confirm whether a command was ran or not.
        location_index: `Literal[1, 2, 3, "random"]`
            The place to commit the crime in.
        Raises
        --------
        InvalidComponent
            invalidcomponentdescription
        NonceTimeout
            Couldn't get the nonce from the Websocket.
        UnknownChannel
            Bot doesn't have access to that channel.

        Returns
        --------
        message: Optional[`Message`]
        """
        _location = None
        if location_index not in [1, 2, 3, "random"]:
            _location = "random"
        _location = location_index if location_index in [1, 2, 3, "random"] else randint(1, 3)
        cmd : Message = self.run_command("crime", retry_attempts, timeout)
        if Parser.check_cooldown(cmd.embeds[0].description):
            return Parser.cooldown(cmd.embeds[0].description)
        self.click(cmd.buttons[_location - 1])
        def check(message):
            return message.author.id == 270904126974590976 and "committed" in message.embeds[0].authorName
        cmd = self.wait_for("MESSAGE_UPDATE", check=check)
        return Parser.crime(cmd.embeds[0].description)

    def postmemes(self, retry_attempts: int = 3, timeout:int = 10, platform:Literal["discord", "reddit", "twitter", "facebook", "random"] = "random", type:Literal["fresh", "repost", "intellectual", "copypasta", "kind", "random"] = "random"):
        """
        Runs the `postmemes` command.
        
        Arguments
        --------
        retry_attempts: `int`
            The amount of times to retry when executing the command fails.
        timeout: `int`
            The time it waits for to confirm whether a command was ran or not.
        platform: `Literal["discord", "reddit", "twitter", "facebook", "random"]`
            The platform to post the meme in.
        type: `Literal["fresh", "repost", "intellectual", "copypasta", "kind", "random"]`
            The type of meme to post.
        Raises
        --------
        InvalidComponent
            invalidcomponentdescription
        NonceTimeout
            Couldn't get the nonce from the Websocket.
        UnknownChannel
            Bot doesn't have access to that channel.

        Returns
        --------
        message: Optional[`Message`]
        """
        _platform = None
        _type = None
        if platform.lower() not in ["discord", "reddit", "twitter", "facebook"]:
            _platform = "random"
        else:
            _platform = ["discord", "reddit", "twitter", "facebook"].index(platform.lower())
        if type.lower() not in ["fresh", "repost", "intellectual", "copypasta", "kind"]:
            _type = "random"
        else:
            _type = ["fresh", "repost", "intellectual", "copypasta", "kind"].index(type.lower())
        _platform = _platform if not _platform == "random" else randint(0, 3)
        _type = _type if not _type == "random" else randint(0, 4)
        cmd : Message = self.run_command("postmemes", retry_attempts, timeout)
        if Parser.check_cooldown(cmd.embeds[0].description):
            return Parser.cooldown(cmd.embeds[0].description)
        def check(message):
            return message.channel_id == int(self.channel_id) and message.author.id == 270904126974590976 and "Meme Posting Session" in message.embeds[0].authorName and "**You" in message.embeds[0].description or "No one" in message.embeds[0].description or "Do you even" in message.embeds[0].description or "Your meme" in message.embeds[0].description or "your meme" in message.embeds[0].description
        self.select(cmd.dropdowns[0], [cmd.dropdowns[0].options[_platform].value])
        self.select(cmd.dropdowns[1], [cmd.dropdowns[1].options[_type].value])
        self.click(cmd.buttons[0])
        cmd = self.wait_for("MESSAGE_UPDATE", check=check)
        return Parser.postmemes(cmd.embeds[0].description)


from typing import Optional, Union
from time import time, sleep
from datetime import datetime, timezone

from requests import post
from requests.exceptions import JSONDecodeError as RequestsJSONDecodeError

from .exceptions import InvalidFormBody
from .Objects import Message, Button, Dropdown

class API:
    """A class that has some useful commands related to communicating with the Discord API."""
    def __init__(self) -> None:
        pass

    def _create_nonce(self) -> str:
        """Creates a nonce using Discord's algorithm.
        
        Returns
        --------
        nonce: str"""
        return str(int(datetime.now(timezone.utc).timestamp() * 1000 - 1420070400000) << 22)

    def _OptionsBuilder(self, name, type_, **kwargs):
        """Builds the data used in slash command API requests.
        
        Parameters
        --------
        name: Any
        type_: Any
        kwargs: **kwargs
        Returns
        --------
        options: list[`dict`[`str`, `Any`]]
        """
        options = [{"type": type_, "name": name, "options": []}]

        type_types = {str: 3, bool: 5, list: 2, int: 10}

        option_type = 3
        for key, value in kwargs.items():
            option_type = type_types[type(value)]
            new_piece_of_data = {"type": option_type, "name": key, "value": value}
            options[0]["options"].append(new_piece_of_data)

        return options

    def _RawOptionsBuilder(self, **kwargs):
        """Builds the raw data used in slash command API requests.
        
        Parameters
        --------
        kwargs: **kwargs
        Returns
        --------
        options: list
        """
        options = []

        type_types = {str: 3, bool: 5, list: 2, int: 10}

        option_type = 3
        for key, value in kwargs.items():
            option_type = type_types[type(value)]
            new_piece_of_data = {"type": option_type, "name": key, "value": value}
            options.append(new_piece_of_data)

        return options

    def run_command(self, name: str, retry_attempts: int = 3, timeout: int = 10, **kwargs) -> Optional[Message]:
        """Runs a slash command.
        Parameters
        --------
        name: str
            The command name.
        retry_attempts: int = 3
            The amount of times to retry on failure.
        timeout: int = 10
            Duration before it times out.
        kwargs: **kwargs
        Returns
        --------
        message: Optional[`Message`]
        """
        nonce = self._create_nonce()
        command_info = self._get_command_info(name) # type: ignore

        retry_attempts = retry_attempts if retry_attempts > 0 else 1
        type_ = 1
        options = []

        if command_info.get("options"):
            for item in command_info["options"]:
                if item["name"] == name:
                    type_ = item["type"]
                    break
            options = self._OptionsBuilder(name, type_, **kwargs)

        data = {
            "type": 2,
            "application_id": "270904126974590976",
            "guild_id": str(self.guild_id), # type: ignore
            "channel_id": self.channel_id, # type: ignore
            "session_id": self.session_id, # type: ignore
            "data": {
                "version": command_info["version"],
                "id": command_info["id"],
                "name": name,
                "type": 1,
                "options": options,
                "application_command": {
                    "id": command_info["id"],
                    "application_id": "270904126974590976",
                    "version": command_info["version"],
                    "default_permission": command_info["default_permission"],
                    "default_member_permissions": command_info["default_member_permissions"],
                    "type": 1,
                    "name": name,
                    "nsfw": command_info["nsfw"],
                    "description": command_info["description"],
                    "dm_permission": command_info["dm_permission"],
                },
                "attachments": [],
            },
            "nonce": nonce,
        }
        def check(message: Message):
            return message.nonce == nonce

        for i in range(retry_attempts):
            response = post(  # type: ignore
                    "https://discord.com/api/v9/interactions",
                    json=data,
                    headers={"Authorization": self.token, "Content-type": "application/json"} # type: ignore
                )
            try:
                response_data = response.json()
                _errors: dict = response_data.get("errors", {}).get("data", {}).get("values", {}).get("0", {}).get("_errors", {})
                if response_data.get("retry_after"):
                    sleep(response_data["retry_after"])
                    continue
                if _errors.get("message"):
                    raise InvalidFormBody(_errors.get("message"))
            except RequestsJSONDecodeError:
                pass
            try:
                interaction: Optional[Union[Message, bool]] = self.wait_for("MESSAGE_CREATE", check=check, timeout=timeout)  # type: ignore
                return interaction # type: ignore
            except Exception as e:
                print(f"Error in run_command: {e}")
                continue
            
        return None

    def run_sub_command(self, name: str, sub_name: str, retry_attempts:int = 3, timeout: int = 10, **kwargs) -> Optional[Message]:
        """Runs a slash command.
        Parameters
        --------
        name: str
            The command name.
        sub_name: str
            The sub command name.
        retry_attempts: int = 3
            The amount of times to retry on failure.
        timeout: int = 10
            Duration before it times out.
        Returns
        --------
        message: Optional[`Message`]
        """

        nonce = self._create_nonce()
        command_info = self._get_command_info(name) # type: ignore
        retry_attempts = retry_attempts if retry_attempts > 0 else 1
        type_ = 1
        options = []

        if command_info.get("options"):
            for item in command_info["options"]:
                if item["name"] == name:
                    type_ = item["type"]
                    break
            options = self._OptionsBuilder(name, type_, **kwargs)

        data = {
            "type": 2,
            "application_id": "270904126974590976",
            "guild_id": str(self.guild_id), # type: ignore
            "channel_id": self.channel_id, # type: ignore
            "session_id": self.session_id, # type: ignore
            "data": {
                "version": command_info["version"],
                "id": command_info["id"],
                "name": name,
                "type": command_info["type"],
                "options": options,
                "application_command": {
                    "id": command_info["id"],
                    "application_id": "270904126974590976",
                    "version": command_info["version"],
                    "default_permission": command_info["default_permission"],
                    "default_member_permissions": command_info["default_member_permissions"],
                    "type": command_info["type"],
                    "name": name,
                    "nsfw": command_info["nsfw"],
                    "description": command_info["description"],
                    "dm_permission": command_info["dm_permission"],
                    "options": command_info["options"],
                },
                "attachments": [],
            },
            "nonce": nonce,
        }
        def check(message: Message):
            return message.nonce == nonce
        for i in range(retry_attempts):
            response = post(  # type: ignore
                    "https://discord.com/api/v9/interactions",
                    json=data,
                    headers={"Authorization": self.token, "Content-type": "application/json"} # type: ignore
                )
            
            try:
                response_data = response.json()
                _errors: dict = response_data.get("errors", {}).get("data", {}).get("values", {}).get("0", {}).get("_errors", {})
                if response_data.get("retry_after"):
                    sleep(response_data["retry_after"])
                    continue
                if _errors.get("message"):
                    raise InvalidFormBody(_errors.get("message"))
            except RequestsJSONDecodeError:
                pass
            try:
                message: Optional[Union[Message, bool]] = self.wait_for("MESSAGE_CREATE", check=check, timeout=timeout) # type: ignore
                return message # type: ignore
            except Exception as e:
                print(f"Error in run_sub_command: {e}")
                continue
        
        return None

    def run_slash_group_command(self, name: str, sub_name: str, sub_group_name: str, retry_attempts: int = 3, timeout: int = 10, **kwargs) -> Optional[Message]:
        """Runs a slash group command.
        Parameters
        --------
        name: str
            The command name.
        sub_name: str
            The sub command name.
        sub_group_name: str
            The sub group command name.
        retry_attempts: int = 3
            The amount of times to retry on failure.
        timeout: int = 10
            Duration before it times out.
        Returns
        --------
        message: Optional[`Message`]
        """

        nonce = self._create_nonce()
        command_info = self._get_command_info(name) # type: ignore
        retry_attempts = retry_attempts if retry_attempts > 0 else 1
        sub_type = 1
        sub_group_type = 1
        options = []

        if command_info.get("options"):
            for item in command_info["options"]:
                if item["name"] == sub_name:
                    sub_type = item["type"]
                for item_ in item["options"]:
                    if item_["name"] == sub_group_name:
                        sub_group_type = item_["type"]
            options = self._RawOptionsBuilder(**kwargs)

        data = {
            "type": 2,
            "application_id": "270904126974590976",
            "guild_id": str(self.guild_id), # type: ignore
            "channel_id": self.channel_id, # type: ignore
            "session_id": self.session_id, # type: ignore
            "data": {
                "version": command_info["version"],
                "id": command_info["id"],
                "name": name,
                "type": command_info["type"],
                "options": [
                    {
                        "type": sub_type,
                        "name": sub_name,
                        "options": [
                            {
                                "type": sub_group_type,
                                "name": sub_group_name,
                                "options": options,
                            }
                        ],
                    }
                ],
                "application_command": {
                    "id": command_info["id"],
                    "application_id": "270904126974590976",
                    "version": command_info["version"],
                    "default_permission": command_info["default_permission"],
                    "default_member_permissions": command_info["default_member_permissions"],
                    "type": command_info["type"],
                    "name": name,
                    "nsfw": command_info["nsfw"],
                    "description": command_info["description"],
                    "dm_permission": command_info["dm_permission"],
                    "options": command_info["options"],
                },
                "attachments": [],
            },
            "nonce": nonce,
        }
        def check(message: Message):
            return message.nonce == nonce

        for i in range(retry_attempts):
            response = post(  # type: ignore
                    "https://discord.com/api/v9/interactions",
                    json=data,
                    headers={"Authorization": self.token, "Content-type": "application/json"} # type: ignore
                )
                
            try:
                response_data = response.json()
                _errors: dict = response_data.get("errors", {}).get("data", {}).get("values", {}).get("0", {}).get("_errors", {})
                if response_data.get("retry_after"):
                    sleep(response_data["retry_after"])
                    continue
                if _errors.get("message"):
                    raise InvalidFormBody(_errors.get("message"))
            except RequestsJSONDecodeError:
                pass
            
            try:
                message: Optional[Union[Message, bool]] = self.wait_for("MESSAGE_CREATE", check=check, timeout=timeout) # type: ignore
                return message # type: ignore
            except Exception as e:
                print(f"Error in run_slash_group_command: {e}")
                continue
        
        return None

    def click(self, button: Button, retry_attempts: int = 10, timeout: int = 10) -> bool:
        """Clicks a button.
        
        Parameters
        --------
        button: Button
            The button to click.
        retry_attempts: int = 10
            The amount of times to retry on failure.
        timeout: int = 10
            Duration before it times out.
        Returns
        --------
        success_state: bool
            Whether the button was clicked successfully or not.
        """
        nonce = self._create_nonce()
        retry_attempts = retry_attempts if retry_attempts > 0 else 1
        data = {
            "type": 3,
            "nonce": nonce,
            "guild_id": str(self.guild_id), # type: ignore
            "channel_id": self.channel_id, # type: ignore
            "message_flags": 0,
            "message_id": button.message_id,
            "application_id": "270904126974590976",
            "session_id": self.session_id, # type: ignore
            "data": {"component_type": 2, "custom_id": button.custom_id},
        }
        end = time() + timeout
        for i in range(retry_attempts):
            if time() > end:
                return False
            response = post(  # type: ignore
                    "https://discord.com/api/v9/interactions",
                    json=data,
                    headers={"Authorization": self.token, "Content-type": "application/json"} # type: ignore
                )
                
            try:
                response_data = response.json()
                _errors: dict = response_data.get("errors", {}).get("data", {})
                try:
                    _errors = _errors[0]["code"]
                    if _errors:
                        raise InvalidFormBody(_errors.get("message"))
                except:
                    pass
            except RequestsJSONDecodeError:
                pass
            if response.status_code == 204:
                break
            retry_after = int(response.json().get('retry_after', 0))
            end += retry_after
            sleep(retry_after)
        
        return True if response.status_code == 204 else False # type: ignore

    def select(self, dropdown: Dropdown, options: list, retry_attempts: int = 10, timeout: int = 10) -> bool:
        """Selects an option from a dropdown.
        Parameters
        dropdown: Dropdown
            The dropdown to choose from.
        options: list
            A list of options to use in the API request for the dropdown to be chosen from.
        retry_attempts: int = 10
            The amount of times to retry on failure.
        timeout: int = 10
        --------
        dropdown: Dropdown
            The dropdown to choose from.
        options: list
            A list of options to use in the API request for the dropdown to be chosen from.
        retry_attempts: int = 10
            The amount of times to retry on failure.
        timeout: int = 10
            Duration before it times out.
        
        Returns
        -------
        success_state: bool
            Whether the option was chosen successfully or not.
        """
        nonce = self._create_nonce()
        data = {
            "type": 3,
            "nonce": nonce,
            "guild_id": str(self.guild_id), # type: ignore
            "channel_id": self.channel_id, # type: ignore
            "message_flags": 0,
            "message_id": dropdown.message_id,
            "application_id": "270904126974590976",
            "session_id": self.session_id, # type: ignore
            "data": {
                "component_type": 3,
                "custom_id": dropdown.custom_id,
                "type": 3,
                "values": options,
            },
        }

        end = time() + timeout
        for i in range(retry_attempts):
            if time() > end:
                return False
            response = post(  # type: ignore
                    "https://discord.com/api/v9/interactions",
                    json=data,
                    headers={"Authorization": self.token, "Content-type": "application/json"} # type: ignore
                )
                
            try:
                response_data = response.json()
                _errors: dict = response_data.get("errors", {}).get("data", {}).get("values", {}).get("0", {}).get("_errors", {})
                if _errors.get("message"):
                    raise InvalidFormBody(_errors.get("message"))
            except RequestsJSONDecodeError:
                pass

            if response.status_code == 204:
                break
            retry_after = int(response.json().get('retry_after', 0))
            end += retry_after
            sleep(retry_after)
        
        return True if response.status_code == 204 else False # type: ignore


import datetime, json, time
import requests

from rich import print
from typing import Callable, Literal, Optional, Union
from pyloggor import pyloggor
from requests import Response

from .exceptions import DataAccessFailure, MissingPermissions, NoCommands, UnknownChannel
from .gateway import Gateway
from .Objects import Config, Message, User
from .core import Core
from .api import API

class Client(API):
    def __init__(self, config: Config, logger: pyloggor):
        __boot_start = time.perf_counter()
        logger.log(level="Info", msg="Booting up DankCord client.")
        self.token = config.token
        self.logger = logger

        self.channel_id: str = str(config.channel_id)
        self.dm_mode = config.dm_mode

        self.resource_intensivity = config.resource_intensivity
        self.commands_data = {}
        self.ws_cache = {}

        self.gateway = Gateway(config, self.logger)
        if not self.gateway:
            raise ConnectionError("Failed to connect to gateway.")

        self.ws = self.gateway.ws
        self.session_id: Optional[str] = self.gateway.session_id

        if not config.dm_mode:
            self.guild_id = self.gateway.guild_id

        self._get_commands()
        self._get_info()

        logger.log(
            level="Info", msg=f"Fully booted up, it took total {round(time.perf_counter() - __boot_start, 3)} seconds."
        )
        self.core = Core(config, self.commands_data, self.guild_id, self.session_id, self.logger, self.gateway)

    def _get_commands(self, channel_id: Optional[str] = None) -> Optional[Response]:
        """Gets all slash command data in a channel, dumps them into memory or a file based on user settings.
        
        Parameters
        ---------
        channel_id: Optional[`str`]
            The channel ID, used for finding a channel and getting slash commands from.

        Raises
        ---------
        UnknownChannel
            Bot doesn't have access to this channel.
        MissingPermissions
            Bot cannot view the channel or read it's content.
        NoCommands
            No commands found.

        Returns
        ---------
        response: Optional[`Response`]
        """
        channel_id = channel_id or self.channel_id
        response = requests.get(  # type: ignore
                f"https://discord.com/api/v9/channels/{channel_id}/application-commands/search?type=1&application_id=270904126974590976",
                headers={"Authorization": self.token, "Content-type": "application/json"},
        )

        if isinstance(response, str):
            return response

        data = response.json()

        if data.get("code") == 10003:
            raise UnknownChannel("Bot doesn't have access to this channel.")
        if data.get("code") == 50013:
            raise MissingPermissions("Bot cannot view the channel or read it's content.")

        if "application_commands" not in data:
            raise NoCommands("No commands found.")

        if self.resource_intensivity == "MEM":
            self.commands_data = {
                command_data["name"]: command_data for command_data in data["application_commands"]
            }
        else:
            with open(f"{self.channel_id}_commands.json", "w") as f:
                json.dump(
                    {command_data["name"]: command_data for command_data in data["application_commands"]},
                    f,
                    indent=4,
                    sort_keys=True,
                )

        return response

    def _get_command_info(self, name: str) -> dict:
        """Retuns information about a given command.
        
        Parameters
        --------
        name: str
            The name of the command.
            
        Returns
        --------
        information: dict
        """
        if self.resource_intensivity == "MEM":
            return self.commands_data.get(name, {})
        else:
            return json.load(open(f"{self.channel_id}_commands.json", "r+")).get(name, {})

    def _get_info(self) -> None:
        """Saves information about the bot user account.
        
        Raises
        -------
        DataAccessFailure
            Failed to get user info.
        """
        resp = requests.get(  # type: ignore
                "https://discord.com/api/v10/users/@me",
                headers={"Authorization": self.token, "Content-type": "application/json"},
            )
        if resp.status_code!= 200:
            raise DataAccessFailure("Failed to get user info.")
        self.user = User(resp.json())

    def wait_for(
        self,
        event: Literal["MESSAGE_CREATE", "MESSAGE_UPDATE", "INTERACTION_CREATE", "INTERACTION_SUCCESS"],
        check: Optional[Callable[..., bool]] = None,
        timeout: float = 10
    ) -> Optional[Union[Message, bool]]:
        """
        Waits for a WebSocket event to be dispatched.

        This could be used to wait for a message to be sent or a message to be edited,
        or even to confirm an interaction being created or successful.

        The `timeout` parameter specifies how long to wait for until the desired event
        is dispatched; if the event was not dispatched before the timeout duration is over,
        it returns `None`.

        This function returns the **first event that meets the requirements**.

        Example
        ---------

        Waiting for a message to be sent:

            def DankMemerShop():
                def check(message: Message):
                    # The author ID is the ID of Dank Memer
                    return message.author.id == 270904126974590976 and "shop" in message.embeds[0].title.lower()
                
                message: Message = bot.wait_for("MESSAGE_CREATE", check = check)

        Parameters
        ------------
        event: str
            The event name.
        check: Optional[Callable[..., `bool`]]
            A predicate to check what to wait for. The arguments must meet the
            parameters of the event being waited for.
        timeout: Optional[`float`]
            The number of seconds to wait before timing out and returning `None`.

        Returns
        --------
        Optional[Union[Message, bool]]
            Returns the Message object, or a boolean.
        """
        limit = time.time() + timeout
        if check is None:
            def _check(*args):
                return True
            check = _check

        while time.time() < limit:
            cache = self.gateway.cache    
            try:        
                if event == "INTERACTION_CREATE":
                    if check(cache.interaction_create[-1]):
                        return True
                if event =="INTERACTION_SUCCESS":
                    if check(cache.interaction_success[-1]):
                        return True

                if event == "MESSAGE_CREATE":
                    value = list(cache.message_create.values())[-1]
                    _msg = Message(value)
                    if check(_msg) is True:
                        return _msg
                if event == "MESSAGE_UPDATE":
                    while len(cache.raw_message_updates) == 0:
                        continue
                    _msg = Message(cache.raw_message_updates[-1])
                    if check(_msg) is True:
                        return _msg
            except IndexError:
                pass
        return None